unit process;

interface

uses SysUtils,FIDUnit,Classes,
   //  StdCtrls,  // for TLabel;
     ComCtrls; // for TProgressBar;




type TProcessMode=(cutHead,cutTail,phaseZero,ZeroFill,
                   FFT,
                   Pivot,phaseFirst,IFFT,
                   DCOffsetCut,linearBaselineCorrection,
                   LT,ILT,
                   AddFID,subtractFID,multiplyFID,divideFID,
                   JoinFID, Flatten, Transpose, averageOverT1,
                   Covariance,CrossCorrelation,
                   phaseCovarianceWeighting,
                   extrapolate,apodization,
                   multiplyScalar, divideScalar,
                   Hausholder, inverseHausholder
                   );





  type TProcessElement=class
    private

    public
      error: boolean;
      errorMessage: string;
      warning: boolean;
      warningMessage: string;
      Mode:TProcessMode;
      function Process(FID:TFID):integer; overload;virtual;abstract;
      function Process(FID2D:TFID_2D):integer;overload;virtual;abstract;
  end;


  type

  TProcessOperations=class
    protected
      FCount: cardinal;

    public
    //  hide: boolean;
      Items: array of TProcessElement;
      constructor create;
      destructor free;
      procedure add(mode:TProcessMode);
      procedure delete(k:cardinal); overload;
      procedure delete; overload;

      property Count: cardinal read FCount;
  end;


  type TRootMatrixThread=class(TThread)

    private
      sqrMat:TProcessElement;

    protected
      procedure Execute; override;

    public
      constructor Create; virtual;

  end;






  TFIDDivideScalar=class(TProcessElement)
  protected
    FScalar: single;
  public
    constructor create;
    function Process(FID:TFID):integer; overload;override;
    function Process(FID2D:TFID_2D):integer; overload;override;
    function caption:string;
    procedure setScalar(a:single);
    property scalar:single read FScalar write setScalar;
  end;


  TFIDMultiplyScalar=class(TProcessElement)
  protected
    FScalar: single;
  public
    constructor create;
    function Process(FID:TFID):integer; overload;override;
    function Process(FID2D:TFID_2D):integer; overload;override;
    function caption:string;
    procedure setScalar(a:single);
    property scalar:single read FScalar write setScalar;
  end;

  (*
  TQL=class(TProcessElement)
  protected

    function tqli(FID2D:TFID_2D):integer;
    function pythag(a,b: double):double;

  public
    constructor create;
    function Process(FID2D:TFID_2D):integer; override;
    function caption:string;
  end;
 *)


  THausholder=class(TProcessElement)
  protected

    function tred2(FID2D:TFID_2D):integer;
    function hausholder_1(FID2D:TFID_2D):integer;


  public
    constructor create;
    function Process(FID2D:TFID_2D):integer; override;
    function hausholder_2(FID2D:TFID_2D;progress:TProgressbar):integer;
 //   function hausholder_2(FID2D:TFID_2D):integer;overload;
    function sqr(FID2D:TFID_2D):integer;
    function caption:string;
  end;

  TinverseHausholder=class(TProcessElement)
  protected

  public
    constructor create;
    function Process(FID2D:TFID_2D):integer; override;
    function caption:string;
  end;


  TCrossCorrelation=class(TProcessElement)
  protected

  public
    constructor create;
    function Process(FID2D:TFID_2D):integer; override;
    function caption:string;
  end;



type TCovarianceMode=(cvReal,cvHyperComplex);

  TCovariance=class(TProcessElement)
  protected
    function hyperComplexCovariance(FID2D:TFID_2D):integer;
    function realCovariance(FID2D:TFID_2D):integer;
  public
    covarianceMode: TCovarianceMode;
    innerProductQ: boolean;
    constructor create;
    function Process(FID2D:TFID_2D):integer; override;
    function caption:string;
  end;

  TPCW=class(TProcessElement)
  protected

  public
    softwarePhaseIncrement:boolean;
    phaseIncrementDeg:single;
    constructor create;
    function Process(FID2D:TFID_2D):integer;override;
    function caption:string;
  end;

  TaverageOverT1=class(TProcessElement)
  protected

  public
    constructor create;
    function Process(FID2D:TFID_2D):integer; override;
    function caption:string;
  end;

  TFlatten=class(TProcessElement)
  protected

  public
    constructor create;
    function Process(FID2D:TFID_2D):integer; override;
    function caption:string;
  end;


  TTranspose=class(TProcessElement)
  protected

  public
    constructor create;
    function Process(FID2D:TFID_2D):integer; override;
    function caption:string;
  end;


  TFIDJoinFID=class(TProcessElement)
  protected
    FParFilename:string;
    FDatFilename:string;
    FWeight:single;
  public
    constructor create; overload;
    constructor create(weight:single); overload;
    function Process(FID2D:TFID_2D):integer; override;
    function caption:string;
    procedure setParFilename(fn:string);
    procedure setDatFilename(fn:string);
    procedure setWeight(w:single);
    property weight:single read FWeight write setWeight;
    property parFilename:string read FParFilename write setParFilename;
    property datFilename:string read FDatFilename write setDatFilename;

  end;


  TFIDAddFID=class(TProcessElement)
  protected
    FParFilename:string;
    FDatFilename:string;
    FWeight:single;
  public
    constructor create;overload;
    constructor create(weight:single);overload;

    function Process(FID2D:TFID_2D):integer; overload;override;
    function caption:string;
    procedure setParFilename(fn:string);
    procedure setDatFilename(fn:string);
    procedure setWeight(w:single);
    property weight:single read FWeight write setWeight;
    property parFilename:string read FParFilename write setParFilename;
    property datFilename:string read FDatFilename write setDatFilename;

  end;


  TFIDSubtractFID=class(TProcessElement)
  protected
    FParFilename:string;
    FDatFilename:string;
    FWeight:single;
  public
    constructor create(weight:single);overload;
    constructor create;overload;
    function Process(FID2D:TFID_2D):integer; override;
    function caption:string;
    procedure setParFilename(fn:string);
    procedure setDatFilename(fn:string);
    procedure setWeight(w:single);
    property weight:single read FWeight write setWeight;
    property parFilename:string read FParFilename write setParFilename;
    property datFilename:string read FDatFilename write setDatFilename;

  end;

  TFIDMultiplyFID=class(TProcessElement)
  protected
    FParFilename:string;
    FDatFilename:string;
  public
    constructor create;
    function Process(FID2D:TFID_2D):integer; override;
    function caption:string;
    procedure setParFilename(fn:string);
    procedure setDatFilename(fn:string);
    property parFilename:string read FParFilename write setParFilename;
    property datFilename:string read FDatFilename write setDatFilename;

  end;

  TFIDDivideFID=class(TProcessElement)
  protected
    FParFilename:string;
    FDatFilename:string;
  public
    constructor create;
    function Process(FID2D:TFID_2D):integer; override;
    function caption:string;
    procedure setParFilename(fn:string);
    procedure setDatFilename(fn:string);
    property parFilename:string read FParFilename write setParFilename;
    property datFilename:string read FDatFilename write setDatFilename;

  end;

  TFIDdcOffsetCut=class(TProcessElement)
  protected
    FPoints:cardinal;   //뉽|Cg̕ςƂĈH
  public
    constructor create;
    function Process(FID:TFID):integer; overload;override;
    function Process(FID2D:TFID_2D):integer; overload;override;
    function caption:string;
    procedure setPoints(p:cardinal);
    property points:cardinal read FPoints write setPoints;
  end;

  TFIDlinearBaselineCorrection=class(TProcessElement)
  protected
    Fx1,Fx2:integer;
    Fy1Re,Fy1Im,Fy2Re,Fy2Im:single;
  public
    constructor create;
    function Process(FID:TFID):integer; overload;override;
    function Process(FID2D:TFID_2D):integer; overload;override;
    function caption:string;
    procedure setInitialPoint(x:integer;yRe,yIm:single);
    procedure setFinalPoint(x:integer;yRe,yIm:single);
    property x1:integer read Fx1;
    property y1Re:single read Fy1Re;
    property y1Im:single read Fy1Im;
    property x2:integer read Fx2;
    property y2Re:single read Fy2Re;
    property y2Im:single read Fy2Im;
  end;

  type TapodizationType=(Gaussian,Lorenzian,Sinc);

  TFIDApodization=class(TProcessElement)
  protected
    FWidth:single; // line width
    FApType:TapodizationType;
  public
    constructor create;
    function Process(FID:TFID):integer; overload;override;
    function Process(FID2D:TFID_2D):integer;overload;override;
    function caption:string;
    procedure setWidth(w:single);
    procedure setApType(ap:TapodizationType);
    property width:single read FWidth write setWidth;
    property apType:TapodizationType read FApType write setApType;
  end;


  TFIDCutHead=class(TProcessElement)
  protected
    FPoints: cardinal;  // |Cg邩H
  public
    constructor create;
    //destructor free;
    function Process(FID:TFID):integer; overload;override;
    function Process(FID2D:TFID_2D):integer; overload;override;
    function caption:string;
    procedure setPoints(p:cardinal);
    property points:cardinal read Fpoints write setPoints;
  end;

  TFIDCutTail=class(TProcessElement)
  protected
    FPoints: cardinal;  // |Cg邩H
  public
    constructor create;
    //destructor free;
    function Process(FID:TFID):integer; overload;override;
    function Process(FID2D:TFID_2D):integer; overload;override;
    function caption:string;
    procedure setPoints(p:cardinal);
    property points:cardinal read FPoints write setPoints;
  end;


  TFIDPhase0=class(TProcessElement) // 0̈ʑ
  protected
    FAngle: double;
  public
    constructor Create;
    function Process(FID:TFID):integer; overload;override;
    function Process(FID2D:TFID_2D):integer; overload;override;
    function caption:string;
    procedure setAngle(ph:double);
    property angle:double read FAngle write setAngle;
  end;

  TFIDPhase1=class(TProcessElement) // 0̈ʑ
  protected
    FAngle: double;
  public
    constructor Create;
    function Process(FID:TFID):integer; overload;override;
    function Process(FID2D:TFID_2D):integer; overload;override;
    function caption:string;
    procedure setAngle(ph:double);
    property angle:double read FAngle write setAngle;
  end;


  TFIDExtrapolate=class(TProcessElement)
  protected
    FAddPoints: cardinal;
    FAveragePoints: cardinal;
  public
    constructor Create;
    function Process(FID:TFID):integer; overload;override;
    function Process(FID2D:TFID_2D):integer; overload;override;
    function caption:string;
    procedure setAddPoints(p:cardinal);
    procedure setAveragePoints(p:cardinal);
    property addPoints:cardinal read FAddPoints write setAddPoints;
    property averagePoints:cardinal read FAveragePoints write setAveragePoints;
  end;



type TZeroFillMode=(zfPrepend,zfAppend);

  TFIDZeroFill=class(TProcessElement)
  protected
    FPoints: cardinal;
  public
    zeroFillMode: TZeroFillMode;
    constructor Create;
    function Process(FID:TFID):integer; overload;override;
    function Process(FID2D:TFID_2D):integer; overload;override;
    function caption:string;
  //  function next2pow(FID:TFID):cardinal;
    procedure setPoints(p:cardinal);
    property points:cardinal read FPoints write setPoints;
  end;


  TFIDLT=class(TProcessElement)
  protected
    FA: single;
  public
    constructor Create;
    function Process(FID:TFID):integer; overload;override;
    function Process(FID2D:TFID_2D):integer; overload;override;
    function caption:string;
    procedure setA(a:single);
    property a:single read FA write setA;
  end;

  TFIDILT=class(TProcessElement)
  protected
    FA: single;
  public
    constructor Create;
    function Process(FID:TFID):integer; overload;override;
    function Process(FID2D:TFID_2D):integer; overload;override;
    function caption:string;
    procedure setA(a:single);
    property a:single read FA write setA;
  end;


  TFIDFFT=class(TProcessElement)
  protected

  public
    constructor Create;
    function Process(FID:TFID):integer; overload;override;
    function Process(FID2D:TFID_2D):integer; overload;override;
    function caption:string;
  end;


  TFIDInverseFFT=class(TProcessElement)
  protected

  public
    constructor Create;
    function Process(FID:TFID):integer; overload;override;
    function Process(FID2D:TFID_2D):integer; overload;override;
    function caption:string;
  end;



  TPivot=class(TProcessElement)
  protected
    FPoint:cardinal;
  public
    constructor Create;
    function Process(FID:TFID):integer; overload;override;
    function Process(FID2D:TFID_2D):integer; overload;override;
    function caption:string;
    procedure setPoint(p:cardinal);
    property point:cardinal read FPoint write setPoint;
  end;

  function next2pow(FID:TFID):cardinal;

implementation

uses main;

//------------------------------------------------------------------------------
function next2pow(FID:TFID):cardinal;
var
  k:cardinal;
begin
  k:=1;
  repeat k:=k*2; until k>FID.AL;
  result:=k;
end;
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
constructor TRootMatrixThread.Create;
begin

  sqrMat:=THausholder.create;

  inherited Create(False);
  FreeOnTerminate := True;

end;
//------------------------------------------------------------------------------
procedure TRootMatrixThread.Execute;
begin
  while (not terminated) and (not finished) do begin
    sqrMat.Process(FID_Proc);
  end; // while;
end;
//------------------------------------------------------------------------------
constructor TProcessOperations.create;
begin
  SetLength(Items,0);
  FCount:=0;
 // hide:=false;
end;
//------------------------------------------------------------------------------
destructor TProcessOperations.free;
begin
  delete;
end;
//------------------------------------------------------------------------------
procedure TProcessOperations.add(mode:TProcessMode);
begin
  Inc(FCount);
  SetLength(Items,Count);
  case mode of
    CutHead: Items[Count-1]:=TFIDCutHead.Create;
    CutTail: Items[Count-1]:=TFIDCutTail.Create;
    phaseZero: Items[Count-1]:=TFIDPhase0.Create;
    ZeroFill: Items[Count-1]:=TFIDZeroFill.Create;
    FFT: Items[Count-1]:=TFIDFFT.Create;
    Pivot: Items[Count-1]:=TPivot.Create;
    phaseFirst: Items[Count-1]:=TFIDPhase1.create;
    DCOFfsetCut: Items[Count-1]:=TFIDDCOffsetCut.create;
    linearBaselineCorrection: Items[Count-1]
                  :=TFIDlinearBaselineCorrection.create;
    LT: Items[Count-1]:=TFIDLT.Create;
    ILT: Items[Count-1]:=TFIDILT.Create;
    AddFID: Items[Count-1]:=TFIDAddFID.Create;
    SubtractFID: Items[Count-1]:=TFIDSubtractFID.create;
    multiplyFID: Items[Count-1]:=TFIDmultiplyFID.Create;
    divideFID: Items[Count-1]:=TFIDdivideFID.create;
    JoinFID: Items[Count-1]:=TFIDJoinFID.Create;
    Flatten: Items[Count-1]:=TFlatten.create;
    Transpose: Items[Count-1]:=TTranspose.create;
    averageOverT1: Items[Count-1]:=TaverageOverT1.create;
    Covariance: Items[Count-1]:=TCovariance.create;
    CrossCorrelation: Items[Count-1]:=TCrossCorrelation.create;
    PhaseCovarianceWeighting: Items[Count-1]:=TPCW.create;
    Extrapolate: Items[Count-1]:=TFIDExtrapolate.create;
    Apodization: Items[Count-1]:=TFIDApodization.create;
    multiplyScalar: Items[Count-1]:=TFIDmultiplyScalar.create;
    divideScalar: Items[Count-1]:=TFIDDivideScalar.Create;
    Hausholder: Items[Count-1]:=THausholder.create;
    inverseHausholder: Items[Count-1]:=TinverseHausholder.create;


  end; // case
end;
//------------------------------------------------------------------------------
procedure TProcessOperations.delete(k:cardinal);
var
  j:integer;
begin
  if k>Count-1 then exit;

  for j:=k to Count-2 do Items[k]:=Items[k+1];

  Items[Count-1].free;
  Dec(FCount);
  SetLength(Items,Count);
end;
//------------------------------------------------------------------------------
procedure TProcessOperations.delete;
begin
  while(Count>0) do delete(Count-1);
end;
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
//------   CONSTRUCTORS   ------------------------------------------------------
//------------------------------------------------------------------------------
constructor TPCW.create;
begin
  error:=false;
  errorMessage:='';
  mode:=phaseCovarianceWeighting;
  softwarePhaseIncrement:=false;
  phaseIncrementDeg:=0.0;
end;

constructor TaverageOverT1.create;
begin
  error:=false;
  errorMessage:='';
  mode:=averageOverT1;
end;

constructor TTranspose.create;
begin
  error:=false;
  errorMessage:='';
  mode:=Transpose;
end;

constructor TFlatten.create;
begin
  error:=false;
  errorMessage:='';
  mode:=Flatten;
end;

constructor TinverseHausholder.create;
begin
  error:=false;
  errorMessage:='';
  mode:=inverseHausholder;
end;

constructor THausholder.create;
begin
  error:=false;
  errorMessage:='';
  mode:=Hausholder;
end;


constructor TCrossCorrelation.create;
begin
  error:=false;
  errorMessage:='';
  mode:=CrossCorrelation;
end;


constructor TCovariance.create;
begin
  error:=false;
  errorMessage:='';
  mode:=Covariance;

  innerProductQ:=false;

  covarianceMode:=cvReal;
end;


constructor TFIDDivideScalar.create;
begin
  error:=false;
  errorMessage:='';
  FScalar:=1;  //default
  mode:=divideScalar;
end;

constructor TFIDmultiplyScalar.create;
begin
  error:=false;
  errorMessage:='';
  FScalar:=1;  //default
  mode:=multiplyScalar;
end;



constructor TFIDSubtractFID.create;
begin
  create(1.0);
end;

constructor TFIDSubtractFID.create(weight:single);
begin
  error:=false;
  errorMessage:='';
  FParFilename:='';
  FDatFilename:='';
  mode:=SubtractFID;
  Fweight:=weight;
end;

procedure TFIDSubtractFID.setWeight(w:single);
begin
  FWeight:=w;
end;


constructor TFIDJoinFID.Create;
begin
  create(1.0);
end;

constructor TFIDJoinFID.Create(weight:single);
begin
  error:=false;
  errorMessage:='';
  FParFileName:='';
  FDatFileName:='';
  mode:=JoinFID;
  FWeight:=weight;
end;

procedure TFIDJoinFID.setWeight(w: Single);
begin
  FWeight:=w;
end;

constructor TFIDAddFID.create;
begin
  create(1.0);
end;

constructor TFIDAddFID.create(weight:single);
begin
  error:=false;
  errorMessage:='';
  FParFilename:='';
  FDatFilename:='';
  mode:=AddFID;
  FWeight:=weight;
end;

procedure TFIDADDFID.setWeight(w: Single);
begin
  FWeight:=w;
end;

constructor TFIDMultiplyFID.create;
begin
  error:=false;
  errorMessage:='';
  FParFilename:='';
  FDatFilename:='';
  mode:=MultiplyFID;
end;

constructor TFIDDivideFID.create;
begin
  error:=false;
  errorMessage:='';
  FParFilename:='';
  FDatFilename:='';
  mode:=DivideFID;
end;

constructor TFIDLT.Create;
begin
  error:=false;
  errorMessage:='';
  FA:=0;
  mode:=LT;
end;

constructor TFIDILT.Create;
begin
  error:=false;
  errorMessage:='';
  FA:=0;
  mode:=LT;
end;

constructor TFIDdcOffsetCut.create;
begin
  error:=false;
  errorMessage:='';
  FPoints:=0;
  mode:=DCOffsetCut;
end;

constructor TFIDlinearBaselineCorrection.create;
begin
  error:=false;
  errorMessage:='';
  Fx1:=0; Fx2:=0;
  Fy1Re:=0.0;Fy1Im:=0.0;Fy2Re:=0.0;Fy2Im:=0.0;
  mode:=linearBaselineCorrection;
end;

constructor TPivot.Create;
begin
  error:=false;
  errorMessage:='';
  mode:=Pivot;
end;

constructor TFIDZeroFill.Create;
begin
  error:=false;
  errorMessage:='';
  zeroFillMode:=zfAppend;
  mode:=ZeroFill;
end;

constructor TFIDExtrapolate.Create;
begin
  error:=false;
  errorMessage:='';
  mode:=Extrapolate;
end;

constructor TFIDApodization.Create;
begin
  error:=false;
  errorMessage:='';
  FApType:=Gaussian;  // Ƃ肠ftHg̓KEVA
  mode:=Apodization;
end;

constructor TFIDPhase0.Create;
begin
  error:=false;
  errorMessage:='';
  mode:=phaseZero;
end;

constructor TFIDPhase1.Create;
begin
  error:=false;
  errorMessage:='';
  mode:=phaseFirst;
end;

constructor TFIDCutHead.create;
begin
  error:=false;
  errorMessage:='';
  mode:=cutHead;
end;
constructor TFIDCutTail.create;
begin
  error:=false;
  errorMessage:='';
  mode:=cutTail;
end;
constructor TFIDFFT.create;
begin
  error:=false;
  errorMessage:='';
  warning:=false;
  warningMessage:='';
  mode:=FFT;
end;
constructor TFIDInverseFFT.create;
begin
  error:=false;
  errorMessage:='';
  warning:=false;
  warningMessage:='';
  mode:=IFFT;
end;


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//--------     SET PARAMS     --------------------------------------------------
//------------------------------------------------------------------------------


procedure TFIDDivideScalar.setScalar(a:single);
begin
  FScalar:=a;
end;

procedure TFIDMultiplyScalar.setScalar(a:single);
begin
  FScalar:=a;
end;

procedure TFIDApodization.setWidth(w: Single);
begin
  FWidth:=w;
end;

procedure TFIDApodization.setApType(ap: TapodizationType);
begin
  FapType:=ap;
end;

procedure TFIDExtrapolate.setAddPoints(p: Cardinal);
begin
  FAddPoints:=p;
end;

procedure TFIDExtrapolate.setAveragePoints(p: Cardinal);
begin
  FAveragePoints:=p;
end;

procedure TFIDSubtractFID.setParFilename(fn: string);
begin
  FParFilename:=fn;
end;

procedure TFIDSubtractFID.setDatFilename(fn: string);
begin
  FDatFilename:=fn;
end;

procedure TFIDAddFID.setParFilename(fn: string);
begin
  FParFilename:=fn;
end;

procedure TFIDAddFID.setDatFilename(fn: string);
begin
  FDatFilename:=fn;
end;

procedure TFIDMultiplyFID.setParFilename(fn: string);
begin
  FParFilename:=fn;
end;

procedure TFIDMultiplyFID.setDatFilename(fn: string);
begin
  FDatFilename:=fn;
end;

procedure TFIDDivideFID.setParFilename(fn: string);
begin
  FParFilename:=fn;
end;

procedure TFIDDivideFID.setDatFilename(fn: string);
begin
  FDatFilename:=fn;
end;

procedure TFIDJoinFID.setParFilename(fn: string);
begin
  FParFilename:=fn;
end;

procedure TFIDJoinFID.setDatFilename(fn: string);
begin
  FDatFilename:=fn;
end;

procedure TFIDLT.setA(a: Single);
begin
  FA:=a;
end;

procedure TFIDILT.setA(a: Single);
begin
  FA:=a;
end;

procedure TFIDlinearBaselineCorrection.setInitialPoint(x: Integer; yRe: Single; yIm: Single);
begin
  FX1:=x;
  FY1Re:=yRe;
  Fy1Im:=yIm;
end;

procedure TFIDlinearBaselineCorrection.setFinalPoint(x: Integer; yRe: Single; yIm: Single);
begin
  FX2:=x;
  FY2Re:=yRe;
  Fy2Im:=yIm;

end;

procedure TFIDdcOffsetCut.setPoints(p: Cardinal);
begin
  FPoints:=p;
end;
//------------------------------------------------------------------------------
procedure TFIDPhase0.setAngle(ph:double);
begin
  FAngle:=ph;
end;
//------------------------------------------------------------------------------
procedure TFIDPhase1.setAngle(ph:double);
begin
  FAngle:=ph;
end;
//------------------------------------------------------------------------------
procedure TFIDCutHead.setPoints(p:cardinal);
begin
  FPoints:=p;
end;
//------------------------------------------------------------------------------
procedure TFIDCutTail.setPoints(p:cardinal);
begin
  FPoints:=p;
end;
//------------------------------------------------------------------------------
procedure TFIDZeroFill.setPoints(p:cardinal);
begin
  FPoints:=p;
end;
//------------------------------------------------------------------------------
procedure TPivot.setPoint(p:cardinal);
begin
  FPoint:=p;
end;
//------------------------------------------------------------------------------
//-------   PROCESS  -----------------------------------------------------------
//------------------------------------------------------------------------------
function TPCW.Process(FID2D: TFID_2D):integer;
var j,k,n:integer;
   // re,im:double;
    psi: array of double;
    averagedPsi,averagedPsiSquared, sigmaPsi: double;

    phi:array of double;
    averagedPhi, averagedPhiSquared, sigmaPhi: double;
    phiPsi,rho,numerator,denominator: double;
begin
  error:=false;

  n:=FID2D.length_2D;
  setLength(psi,n);

  averagedPsi:=0.0;
  averagedPsiSquared:=0.0;

  if softwarePhaseIncrement=true then begin
    for j:=0 to n-1 do begin
      psi[j]:= (2.0*3.14159265358979*j/n)-3.14159265358979;
                  // 360.0*j/n-180.0
      FID2D.FID[j].rotate(-1.0*psi[j]*180.0/3.14159256358979);
      psi[j]:=abs(psi[j])-(0.5*3.1415926535);
      averagedPsi:=averagedPsi+psi[j];
      averagedPsiSquared:=averagedPsiSquared+(psi[j]*psi[j]);
    end;
  end else begin
    for j:=0 to n-1 do begin
      psi[j]:=j*phaseIncrementDeg*3.14159265358979/180.0-3.14159265358979;
      FID2D.FID[j].rotate(-180.0);
      psi[j]:=abs(psi[j])-(0.5*3.1415926535);
      averagedPsi:=averagedPsi+psi[j];
      averagedPsiSquared:=averagedPsiSquared+(psi[j]*psi[j]);
    end;
  end;

  averagedPsi:=averagedPsi/n;
  averagedPsiSquared:=averagedPsiSquared/n;
  sigmaPsi:=sqrt(averagedPsiSquared - (averagedPsi*averagedPsi));

  setLength(phi,n);
  for k:=0 to FID2D.al-1 do begin
    averagedPhi:=0.0;
    averagedPhiSquared:=0.0;
    for j:=0 to n-1 do begin
      phi[j]:=FID2D.FID[j].phase(k);
      phi[j]:=abs(phi[j])-(0.5*3.1415926535);
      averagedPhi:=averagedPhi+phi[j];
      averagedPhiSquared:=averagedPhiSquared+(phi[j]*phi[j]);
    end;//j
    averagedPhi:=averagedPhi/n;
    averagedPhiSquared:=averagedPhiSquared/n;
    sigmaPhi:=sqrt(averagedPhiSquared - (averagedPhi*averagedPhi));

    phiPsi:=0.0;
    for j:=0 to n-1 do begin
      phiPsi:=phiPsi+phi[j]*psi[j];
    end;
    phiPsi:=phiPsi/n;

    numerator:=phiPsi-averagedPhi*averagedPsi;
    denominator:=sigmaPhi*sigmaPsi;

    rho:=numerator/denominator;
    FID2D.FID[0].real.sig[k]:=rho;
    FID2D.FID[0].imag.sig[k]:=0.0;
    FID2D.FID[0].abs.sig[k]:=sqrt(rho*rho);

  end; // k


  while FID2D.length_2D>1 do begin
    FID2D.delete(FID2D.length_2D-1);
  end;

  result:=0;
end;
//------------------------------------------------------------------------------

function TaverageOverT1.Process(FID2D:TFID_2D):integer;
var j,k:integer;
    re,im: double;
begin
  error:=false;
  for k:=0 to FID2D.al-1 do begin
    for j:=1 to FID2D.length_2D-1 do begin
      FID2D.FID[0].real.sig[k]:=FID2D.FID[0].real.sig[k]+
                               +FID2D.FID[j].real.sig[k];
      FID2D.FID[0].imag.sig[k]:=FID2D.FID[0].imag.sig[k]+
                               +FID2D.FID[j].imag.sig[k];
    end;

  re:=FID2D.FID[0].real.sig[k];
  im:=FID2D.FID[0].imag.sig[k];
  FID2D.FID[0].real.sig[k]:=re/FID2D.length_2D;
  FID2D.FID[0].imag.sig[k]:=im/FID2D.length_2D;
  FID2D.FID[0].abs.sig[k]:=sqrt(re*re + im*im)/FID2D.length_2D;
  end; // k

  while FID2D.length_2D>1 do begin
    FID2D.delete(FID2D.length_2D-1);
  end;

  result:=0;

end;
//------------------------------------------------------------------------------
function TFlatten.Process(FID2D:TFID_2D):integer;
var j,k:integer;
    re,im: double;
begin
  error:=false;
  for j:=0 to FID2D.length_2D-1 do begin
    for k := 1 to FID2D.al-1 do begin
      FID2D.FID[j].real.sig[0]:=FID2D.FID[j].real.sig[0]
                               +FID2D.FID[j].real.sig[k];
      FID2D.FID[j].imag.sig[0]:=FID2D.FID[j].imag.sig[0]
                               +FID2D.FID[j].imag.sig[k];
    end; // k

    FID2D.FID[j].real.sig[0]:=FID2D.FID[j].real.sig[0]/FID2D.al;
    FID2D.FID[j].imag.sig[0]:=FID2D.FID[j].imag.sig[0]/FID2D.al;

    re:=FID2D.FID[j].real.sig[0];
    im:=FID2D.FID[j].imag.sig[0];
    FID2D.FID[j].abs.sig[0]:=sqrt(re*re + im*im);

  end;

  FID2D.al:=1;

  result:=0;

end;
//------------------------------------------------------------------------------
function TFIDJoinFID.Process(FID2D: TFID_2D):integer;
var origAL: integer;
    i,j, k:integer;
    s,s2:string;
    F1:TextFile;
    al:cardinal;
    nFID2D:TFID_2D;
    found:boolean;
begin
  error:=false; al:=0;

  if not fileExists(parFilename) then begin
    error:=true;
    errorMessage:= 'JoinFID: File ' + parFilename + ' does not exist.';
    result:=1;
    exit;
  end;

  if not fileExists(datFilename) then begin
    error:=true;
    errorMessage:= 'JoinFID: File ' + datFilename + ' does not exist.';
    result:=1;
    exit;
  end;

  // read parameter file
  AssignFile(F1, parFilename);
  Reset(F1);
  found:=false;
  repeat
    Readln(F1,s);
    { 󔒂}
    s:=Trim(LowerCase(s));
    s2:='';
    for k:=1 to Length(s) do begin
      if s[k]<>' ' then s2:=s2+s[k];
    end;
    s:=s2;
    if Copy(s,1,6)='point=' then begin
      s2:=Copy(s,7,Length(s));
      al:=StrToInt(s2);
      found:=true;
      break;
    end;
  until EOF(F1);

  CloseFile(F1);
  if not found then begin
    error:=true;
    errorMessage:= 'JoinFID: Entry "point" could not be found in '
       + parFilename + '.';
    result:=1;
    exit;
  end;


  nFID2D:=TFID_2D.create(1.0,al);   // dw͓Kł悢

  nFID2D.Readsm2dFile(datFilename);


  if nFID2D.length_2D <> FID2D.length_2D then begin
    error:=true;
    errorMessage:= 'JoinFID: Cannot join data with different 2D-data lengths.';
    nFID2D.free;
    result:=1;
    exit;
  end;

  origAL:=FID2D.al;
  FID2D.al:=FID2D.al+nFID2D.al;

  for j:=0 to FID2D.length_2D-1 do begin
    for i:=origAL to FID2D.al-1 do begin
      FID2D.FID[j].real.sig[i] := FWeight * nFID2D.FID[j].real.sig[i-origAL];
      FID2D.FID[j].imag.sig[i] := FWeight * nFID2D.FID[j].imag.sig[i-origAL];
      FID2D.FID[j].abs.sig[i]:=sqrt(
           FID2D.FID[j].real.sig[i]*FID2D.FID[j].real.sig[i]
          +FID2D.FID[j].imag.sig[i]*FID2D.FID[j].imag.sig[i]
          );
    end;
  end; // j;

  nFID2D.free;
  result:=0;

end;

//------------------------------------------------------------------------------
function TFIDAddFID.Process(FID2D: TFID_2D):integer;
var i,k:integer;
    s,s2:string;
    found:boolean;
    F1:TextFile;
    al:cardinal;
    nFID2D:TFID_2D;
    j:cardinal;
begin
  error:=false; al:=0;

  if not fileExists(parFilename) then begin
    error:=true;
    errorMessage:= 'AddFID: File ' + parFilename + ' does not exist.';
    result:=1;
    exit;
  end;

  if not fileExists(datFilename) then begin
    error:=true;
    errorMessage:= 'AddFID: File ' + datFilename + ' does not exist.';
    result:=1;
    exit;
  end;

  // read parameter file
  AssignFile(F1, parFilename);
  Reset(F1);
  found:=false;
  repeat
    Readln(F1,s);
    { 󔒂}
    s:=Trim(LowerCase(s));
    s2:='';
    for k:=1 to Length(s) do begin
      if s[k]<>' ' then s2:=s2+s[k];
    end;
    s:=s2;
    if Copy(s,1,6)='point=' then begin
      s2:=Copy(s,7,Length(s));
      al:=StrToInt(s2);
      found:=true;
      break;
    end;
  until EOF(F1);

  CloseFile(F1);
  if not found then begin
    error:=true;
    errorMessage:= 'AddFID: Entry "point" could not be found in '
       + parFilename + '.';
    result:=1;
    exit;
  end;


  if al<>FID2D.al then begin
    error:=true;
    errorMessage:= 'AddFID: Cannot add data with different data lengths.';
    result:=1;
    exit;
  end;


  nFID2D:=TFID_2D.create(1.0,al);   // dw͓Kł悢


  nFID2D.Readsm2dFile(datfilename);

  if nFID2D.length_2D <> FID2D.length_2D then begin
    error:=true;
    errorMessage:= 'AddFID: Cannot add data with different 2D-data lengths.';
    nFID2D.free;
    result:=1;
    exit;
  end;


  for j:=0 to FID2D.length_2D-1 do begin
    for i:=0 to FID2D.al-1 do begin
      FID2D.FID[j].real.sig[i] := FID2D.FID[j].real.sig[i]
                                 + FWeight * nFID2D.FID[j].real.sig[i];
      FID2D.FID[j].imag.sig[i] := FID2D.FID[j].imag.sig[i]
                                 + FWeight * nFID2D.FID[j].imag.sig[i];
      FID2D.FID[j].abs.sig[i]:=sqrt(
           FID2D.FID[j].real.sig[i]*FID2D.FID[j].real.sig[i]
          +FID2D.FID[j].imag.sig[i]*FID2D.FID[j].imag.sig[i]
          );
    end;
  end; // j;

  nFID2D.free;
  result:=0;

end;

//------------------------------------------------------------------------------
function TFIDSubtractFID.Process(FID2D: TFID_2D):integer;
var i,k:integer;
    s,s2:string;
    F1:TextFile;
    al:cardinal;
    nFID2D:TFID_2D;
    found:boolean;
    j:cardinal;
begin
  error:=false; al:=0;

  if not fileExists(parFilename) then begin
    error:=true;
    errorMessage:= 'SubtractFID: File ' + parFilename + ' does not exist.';
    result:=1;
    exit;
  end;

  if not fileExists(datFilename) then begin
    error:=true;
    errorMessage:= 'SubtractFID: File ' + datFilename + ' does not exist.';
    result:=1;
    exit;
  end;

  // read parameter file
  AssignFile(F1, parFilename);
  Reset(F1);
  found:=false;
  repeat
    Readln(F1,s);
    { 󔒂}
    s:=Trim(LowerCase(s));
    s2:='';
    for k:=1 to Length(s) do begin
      if s[k]<>' ' then s2:=s2+s[k];
    end;
    s:=s2;
    if Copy(s,1,6)='point=' then begin
      s2:=Copy(s,7,Length(s));
      al:=StrToInt(s2);
      found:=true;
      break;
    end;
  until EOF(F1);

  CloseFile(F1);
  if not found then begin
    error:=true;
    errorMessage:= 'SubtractFID: Entry "point" could not be found in '
       + parFilename + '.';
    result:=1;
    exit;
  end;


  if al<>FID2D.al then begin
    error:=true;
    errorMessage:= 'SubtractFID: Cannot subtract data with different data lengths.';
    result:=1;
    exit;
  end;


  nFID2D:=TFID_2D.create(1.0,al);   // dw͓Kł悢

  nFID2D.Readsm2dFile(datfilename);

  if nFID2D.length_2D <> FID2D.length_2D then begin
    error:=true;
    errorMessage:= 'SubtractFID: Cannot subtract data with different 2D-data lengths.';
    nFID2D.free;
    result:=1;
    exit;
  end;


  for j:=0 to FID2D.length_2D-1 do begin
    for i:=0 to FID2D.al-1 do begin
      FID2D.FID[j].real.sig[i] := FID2D.FID[j].real.sig[i]
                                 - FWeight * nFID2D.FID[j].real.sig[i];
      FID2D.FID[j].imag.sig[i] := FID2D.FID[j].imag.sig[i]
                                 - FWeight * nFID2D.FID[j].imag.sig[i];
      FID2D.FID[j].abs.sig[i]:=sqrt(
           FID2D.FID[j].real.sig[i]*FID2D.FID[j].real.sig[i]
          +FID2D.FID[j].imag.sig[i]*FID2D.FID[j].imag.sig[i]
          );
    end;
  end; // j;

  nFID2D.free;
  result:=0;

end;

//------------------------------------------------------------------------------

function TFIDMultiplyFID.Process(FID2D: TFID_2D):integer;
var i,k:integer;
    s,s2:string;
    F1:TextFile;
    al:cardinal;
    nFID2D:TFID_2D;
    found:boolean;
    j:cardinal;
    re,im,nre,nim:single;
begin
  error:=false; al:=0;

  if not fileExists(parFilename) then begin
    error:=true;
    errorMessage:= 'MultiplyFID: File ' + parFilename + ' does not exist.';
    result:=1;
    exit;
  end;

  if not fileExists(datFilename) then begin
    error:=true;
    errorMessage:= 'MultiplyFID: File ' + datFilename + ' does not exist.';
    result:=1;
    exit;
  end;

  // read parameter file
  AssignFile(F1, parFilename);
  Reset(F1);
  found:=false;
  repeat
    Readln(F1,s);
    { 󔒂}
    s:=Trim(LowerCase(s));
    s2:='';
    for k:=1 to Length(s) do begin
      if s[k]<>' ' then s2:=s2+s[k];
    end;
    s:=s2;
    if Copy(s,1,6)='point=' then begin
      s2:=Copy(s,7,Length(s));
      al:=StrToInt(s2);
      found:=true;
      break;
    end;
  until EOF(F1);

  CloseFile(F1);
  if not found then begin
    error:=true;
    errorMessage:= 'MultiplyFID: Entry "point" could not be found in '
       + parFilename + '.';
    result:=1;
    exit;
  end;


  if al<>FID2D.al then begin
    error:=true;
    errorMessage:= 'MultiplyFID: Cannot multiply data with different data lengths.';
    result:=1;
    exit;
  end;


  nFID2D:=TFID_2D.create(1.0,al);   // dw͓Kł悢

  nFID2D.Readsm2dFile(datfilename);

  if nFID2D.length_2D <> FID2D.length_2D then begin
    error:=true;
    errorMessage:= 'MultiplyFID: Cannot multiply data with different 2D-data lengths.';
    nFID2D.free;
    result:=1;
    exit;
  end;


  for j:=0 to FID2D.length_2D-1 do begin
    for i:=0 to FID2D.al-1 do begin
      re:=FID2D.FID[j].real.sig[i];
      im:=FID2D.FID[j].imag.sig[i];
      nre:=nFID2D.FID[j].real.sig[i];
      nim:=nFID2D.FID[j].imag.sig[i];
      // complex multiplication
      FID2D.FID[j].real.sig[i] := (re*nre)-(im*nim);
      FID2D.FID[j].imag.sig[i] := (im*nre)+(re*nim);
      FID2D.FID[j].abs.sig[i]:=sqrt(
           FID2D.FID[j].real.sig[i]*FID2D.FID[j].real.sig[i]
          +FID2D.FID[j].imag.sig[i]*FID2D.FID[j].imag.sig[i]
          );
    end;
  end; // j;

  nFID2D.free;
  result:=0;

end;
//------------------------------------------------------------------------------

function TFIDDivideFID.Process(FID2D: TFID_2D):integer;
var i,k:integer;
    s,s2:string;
    F1:TextFile;
    al:cardinal;
    nFID2D:TFID_2D;
    found:boolean;
    j:cardinal;
    re,im,nre,nim,nab:single;
begin
  error:=false; al:=0;

  if not fileExists(parFilename) then begin
    error:=true;
    errorMessage:= 'DivideFID: File ' + parFilename + ' does not exist.';
    result:=1;
    exit;
  end;

  if not fileExists(datFilename) then begin
    error:=true;
    errorMessage:= 'DivideFID: File ' + datFilename + ' does not exist.';
    result:=1;
    exit;
  end;

  // read parameter file
  AssignFile(F1, parFilename);
  Reset(F1);
  found:=false;
  repeat
    Readln(F1,s);
    { 󔒂}
    s:=Trim(LowerCase(s));
    s2:='';
    for k:=1 to Length(s) do begin
      if s[k]<>' ' then s2:=s2+s[k];
    end;
    s:=s2;
    if Copy(s,1,6)='point=' then begin
      s2:=Copy(s,7,Length(s));
      al:=StrToInt(s2);
      found:=true;
      break;
    end;
  until EOF(F1);

  CloseFile(F1);
  if not found then begin
    error:=true;
    errorMessage:= 'DivideFID: Entry "point" could not be found in '
       + parFilename + '.';
    result:=1;
    exit;
  end;


  if al<>FID2D.al then begin
    error:=true;
    errorMessage:= 'DivideFID: Cannot divide data with different data lengths.';
    result:=1;
    exit;
  end;


  nFID2D:=TFID_2D.create(1.0,al);   // dw͓Kł悢

  nFID2D.Readsm2dFile(datfilename);

  if nFID2D.length_2D <> FID2D.length_2D then begin
    error:=true;
    errorMessage:= 'DivideFID: Cannot divide data with different 2D-data lengths.';
    nFID2D.free;
    result:=1;
    exit;
  end;

  for j:=0 to nFID2D.length_2D-1 do begin
    for i:=0 to nFID2D.al-1 do begin
      if nFID2D.FID[j].abs.sig[i]=0 then begin //[ZG[
        error:=true;
        errorMessage:= 'DivideFID: Cannot divide by zero. ('
            + IntToStr(i) + 'th point)';
        nFID2D.free;
        result:=1;
        exit;
      end;
    end; // i;
  end; // j

  for j:=0 to FID2D.length_2D-1 do begin
    for i:=0 to FID2D.al-1 do begin
      re:=FID2D.FID[j].real.sig[i];
      im:=FID2D.FID[j].imag.sig[i];
      nre:=nFID2D.FID[j].real.sig[i];
      nim:=nFID2D.FID[j].imag.sig[i];
      nab:=nFID2D.FID[j].abs.sig[i];

      //complex division
      FID2D.FID[j].real.sig[i] := (re*nre+im*nim)/(nab*nab);
      FID2D.FID[j].imag.sig[i] := (im*nre-re*nim)/(nab*nab) ;

      FID2D.FID[j].abs.sig[i]:=sqrt(
           FID2D.FID[j].real.sig[i]*FID2D.FID[j].real.sig[i]
          +FID2D.FID[j].imag.sig[i]*FID2D.FID[j].imag.sig[i]
          );
    end;
  end; // j;

  nFID2D.free;
  result:=0;

end;
//------------------------------------------------------------------------------
function TFIDExtrapolate.Process(FID2D: TFID_2D):integer;
var k:integer;
begin
  for k:=0 to FID2D.length_2D-1 do begin
    if Process(FID2D.FID[k])<>0 then begin
      result:=1;
      exit;
    end;
  end;

  FID2D.al:=FID2D.FID[0].al;
  result:=0;
end;
//------------------------------------------------------------------------------
function TFIDExtrapolate.Process(FID: TFID):integer;
var k,oldAL:cardinal;
    reo,imo,abo: double;
begin
  if averagePoints>FID.AL then begin
    error:=true;
    errorMessage:='Extrapolate: Too many averaging points.';
    result:=1;
    exit;
  end;

  if FID.AL+Addpoints>65536 then begin
    error:=true;
    errorMessage:='Extrapolate: Too many adding points.';
    result:=1;
    exit;
  end;

  k:=FID.AL-averagePoints;

  reo:=0.0; imo:=0.0;
  while k<FID.AL do begin
    reo:=reo+FID.real.sig[k];
    imo:=imo+FID.imag.sig[k];
    inc(k);
  end;

  if averagePoints>0 then begin
    reo:=reo/averagePoints;
    imo:=imo/averagePoints;
  end;
  abo:=sqrt(reo*reo+imo*imo);

  oldAL:=FID.AL;;
  FID.AL:=oldAL+addPoints;

  for k:=oldAL to FID.AL-1 do begin
    FID.real.sig[k]:=reo;
    FID.imag.sig[k]:=imo;
    FID.abs.sig[k]:=abo;
  end;

  result:=0;

end;

//------------------------------------------------------------------------------
function TFIDdcOffsetCut.Process(FID2D: TFID_2D):integer;
var k:cardinal;
    erra:boolean;
begin

  erra:=false;
  for k:=0 to FID2D.length_2D-1 do begin
    if Process(FID2D.FID[k])<>0 then erra:=true;
  end;

  if erra then result:=1 else result:=0;
end;
//------------------------------------------------------------------------------
function TFIDDivideScalar.Process(FID2D: TFID_2D):integer;
var k:cardinal;
    erra: boolean;
begin
  erra:=false;
  for k:=0 to FID2D.length_2D-1 do begin
     if Process(FID2D.FID[k])<>0 then erra:=true;
  end;
  if erra then result:=1 else result:=0;
end;

//------------------------------------------------------------------------------
function TFIDDivideScalar.Process(FID: TFID):integer;
var k:cardinal;
    reo,imo: double;
begin
  if FScalar=0 then begin
    error:=true;
    errorMessage:='Cannot divide by 0.';
    result:=1;
    exit;
  end;

  for k:=0 to FID.AL-1 do begin
    FID.real.sig[k]:=FID.real.sig[k]/FScalar;
    FID.imag.sig[k]:=FID.imag.sig[k]/FScalar;
    reo:=FID.real.sig[k];
    imo:=FID.imag.sig[k];
    FID.abs.sig[k]:=sqrt(reo*reo+imo*imo);
  end;
  result:=0;
end;

//------------------------------------------------------------------------------
function TFIDmultiplyScalar.Process(FID2D: TFID_2D):integer;
var k:cardinal;
    erra: boolean;
begin
  erra:=false;
  for k:=0 to FID2D.length_2D-1 do begin
     if Process(FID2D.FID[k])<>0 then erra:=true;
  end;
  if erra then result:=1 else result:=0;

end;
//------------------------------------------------------------------------------
function TFIDmultiplyScalar.Process(FID: TFID):integer;
var k:cardinal;
    reo,imo: double;
begin
  for k:=0 to FID.AL-1 do begin
    FID.real.sig[k]:=FID.real.sig[k]*FScalar;
    FID.imag.sig[k]:=FID.imag.sig[k]*FScalar;
    reo:=FID.real.sig[k];
    imo:=FID.imag.sig[k];
    FID.abs.sig[k]:=sqrt(reo*reo+imo*imo);
  end;
  result:=0;
end;
//------------------------------------------------------------------------------
function TFIDApodization.Process(FID2D: TFID_2D):integer;
var k:cardinal;
    erra: boolean;
begin
  erra:=false;
  for k:=0 to FID2D.length_2D-1 do begin
     if Process(FID2D.FID[k])<>0 then erra:=true;
  end;
  if erra then result:=1 else result:=0;
end;
//------------------------------------------------------------------------------
function TFIDApodization.Process(FID: TFID):integer;
var k:cardinal;
    w,reo,imo: double;
begin
  w:=0;

  if width<0 then begin
    error:=true;
    errorMessage:='Apodization: width must be positive';
    result:=1;
  end else begin
    for k:=0 to FID.AL-1 do begin
      case apType of
        Gaussian: w:=exp(-k*k*(FID.dw/1000000.0)*(FID.dw/1000000.0)*width*width*3.55971);
                                    // (2pi) ^2/(16 ln2) = 3.55971
        Lorenzian: w:=exp(-k*(FID.dw/1000000.0)*3.141592*width);

        Sinc: if k=0 then w:=1 else
               w:=sin(k*(FID.dw/1000000.0)*3.141592*width)/
                     (k*(FID.dw/1000000.0)*3.141592*width);
      end;
      FID.real.sig[k]:=FID.real.sig[k]*w;
      FID.imag.sig[k]:=FID.imag.sig[k]*w;
      reo:=FID.real.sig[k];
      imo:=FID.imag.sig[k];
      FID.abs.sig[k]:=sqrt(reo*reo+imo*imo);
    end;
    result:=0;
  end;

end;
//------------------------------------------------------------------------------
function TFIDlinearBaselineCorrection.Process(FID2D: TFID_2D):integer;
var k:cardinal;
    erra:boolean;
begin

  erra:=false;
  for k:=0 to FID2D.length_2D-1 do begin
    if Process(FID2D.FID[k])<>0 then erra:=true;
  end;

  if erra then result:=1 else result:=0;
end;

//------------------------------------------------------------------------------
function TFIDlinearBaselineCorrection.Process(FID: TFID):integer;
var k:integer;
    slopeRe,slopeIm:single;
begin
  if (x1<0) or (x2<0) or (x1>integer(FID.AL)-1) or (x2>integer(FID.AL)-1) then begin
    error:=true;
    errorMessage:='linear baseline correction: Out of range.';
    result:=1;
    exit;
  end;

  if x2-x1=0 then begin
    error:=true;
    errorMessage:='linear baseline correction: x1 and x2 cannot be the same value';
    result:=1;
    exit;
  end;

  if x2<=x1 then begin
    error:=true;
    errorMessage:='linear baseline correction: x2 should be larger than x1';
    result:=1;
    exit;
  end;


  slopeRe:=(y2Re-y1Re)/(x2-x1);
  slopeIm:=(y2Im-y1Im)/(x2-x1);

  for k:=x1 to x2 do begin
    FID.real.sig[k]:=FID.real.sig[k]-(y1Re+slopeRe*(k-x1));
    FID.imag.sig[k]:=FID.imag.sig[k]-(y1Im+slopeIm*(k-x1));
    FID.abs.sig[k]:=sqrt( FID.real.sig[k]*FID.real.sig[k] +
                         FID.imag.sig[k]*FID.imag.sig[k]);
  end;

  result:=0;

end;
//------------------------------------------------------------------------------
function TFIDdcOffsetCut.Process(FID: TFID):integer;
var k:cardinal;
    reo,imo: double;
begin
  if points>FID.AL then begin
    error:=true;
    errorMessage:='DC offset cut: Too many points.';
    result:=1;
    exit;
  end;

  k:=FID.AL-points;

  reo:=0.0; imo:=0.0;
  while k<FID.AL do begin
    reo:=reo+FID.real.sig[k];
    imo:=imo+FID.imag.sig[k];
    inc(k);
  end;

  if points>0 then begin
    reo:=reo/points;
    imo:=imo/points;
  end;

  for k:=0 to FID.AL-1 do begin
    FID.real.sig[k]:=FID.real.sig[k]-reo;
    FID.imag.sig[k]:=FID.imag.sig[k]-imo;
    FID.abs.sig[k]:=sqrt(FID.real.sig[k]*FID.real.sig[k]
                        + FID.imag.sig[k]*FID.imag.sig[k]);
  end;

  result:=0;


end;

//------------------------------------------------------------------------------
function THausholder.tred2(FID2D:TFID_2D):integer;
(*
    Y1 to n ƂBi0 to n-1ł͂ȂIj
    炩SetLength(n+1)ƂĂI

*)
var

  a,b,q:array of array of double;
  n:integer;
  u,p:array of double;

  l,k,j,i:integer;
  e,largeK, H, sqs, scale:double;
begin

  n:=FID2D.al;

  setLength(a,n+1);
  setLength(b,n+1);
  setLength(q,n+1);

  for i:=0 to n do begin
    setLength(a[i],n+1);
    setLength(b[i],n+1);
    setLength(q[i],n+1);
  end;

  setLength(u,n+1);
  setLength(p,n+1);

  for i:=1 to n do for j:=1 to n do a[i][j]:=FID2D.FID[i-1].real.sig[j-1];


  //
  //
  //  Hausholder
  //
  //


  for j:=1 to n do for k:=1 to n do if j=k then q[j][k]:=1.0 else q[j][k]:=0.0;
  //Pʍs

  for i:=n downto 2 do begin

    //
    // scale
    //

    scale:=0.0;
    for k:=1 to i-1 do scale:=scale+abs(a[k][i]);
    if scale=0 then break;

    //
    //  sqrt(sigma)
    //
    sqs:=0.0;
    for k:=1 to i-1 do begin
      a[k][i]:=a[k][i]/scale;
      sqs:=sqs+a[k][i]*a[k][i];
    end; // k
    sqs:=sqrt(sqs);

    //
    //  u
    //
    for k:=1 to i-2 do u[k]:=a[k][i];

    if a[i-1][i]>=0 then begin
      u[i-1]:=a[i-1][i]+sqs;
      e:=-sqs*scale;
    end else begin
      u[i-1]:=a[i-1][i]-sqs;
      e:=sqs*scale;
    end;

    for k:=i to n do u[k]:=0.0;

    //
    //  H
    //
    H:=0.0;
    for k:=1 to i-1 do H:=H+u[k]*u[k];
    H:=H/2.0;
    if H=0 then break;  // scale=0̂ƂbreakĂ̂ŁAH=0ƂȂ邱Ƃ
                        // Ȃ͂AÔ߁AŃ[ŊZȂ悤ɁBBB

    //
    //  p
    //
    for j:=1 to i do begin
      p[j]:=0.0;
      for k:=1 to i-1 do p[j]:=p[j]+a[j][k]*u[k];
      p[j]:=p[j]/H;
    end;

    //
    //  K(largeK)
    //
    largeK:=0.0;
    for k:=1 to i-1 do largeK:=largeK+u[k]*p[k];
    largeK:=largeK/H;

    //
    //   update a[j][k]
    //
    for j:=1 to i-1 do begin
      for k:=1 to i-1 do begin
        a[j][k]:=a[j][k]-p[j]*u[k]-u[j]*p[k]+largeK*u[j]*u[k];
      end;
    end; // j

    for k:=1 to i-2 do begin
      a[k][i]:=0.0;
      a[i][k]:=0.0;
    end;

    a[i-1][i]:= e;
    a[i][i-1]:= e;

    for j:=1 to i-1 do begin
      for k:=1 to i-1 do begin
        b[j][k]:=0.0;
        for l:=1 to i do begin
          if j=l then b[j][k]:=b[j][k]+(1.0-u[j]*u[l]/H)*q[l][k]
          else b[j][k]:=b[j][k]+(0.0-u[j]*u[l]/H)*q[l][k];
        end;
      end;
    end;

    for j:=1 to i-1 do begin
      for k:=1 to i-1 do begin
        q[j][k]:=b[j][k];
      end;
    end;

  end; // i

  //
  //
  //

  for i:=1 to n do for j:=1 to n do begin
    FID2D.FID[i-1].real.sig[j-1]:=a[i][j];
    FID2D.FID[i-1].imag.sig[j-1]:=q[i][j];
    FID2D.FID[i-1].abs.sig[j-1]:=0.0;
  end;

  result:=0;

end;
//------------------------------------------------------------------------------
function THausholder.sqr(FID2D:TFID_2D):integer;
var
  i,j,k,n:integer;
  b:array of array of double;
begin

  n:=FID2D.length_2D;
  setLength(b,n);
  for i:=0 to n-1 do setLength(b[i],n);

  for i:=0 to n-1 do begin
    for j:=0 to n-1 do begin
      b[i][j]:=0.0;
      for k:=0 to n-1 do
        b[i][j]:=b[i][j]+FID2D.FID[i].real.sig[k]*FID2D.FID[k].real.sig[j];
    end;
  end;

  for i:=0 to n-1 do for j:=0 to n-1 do FID2D.FID[i].real.sig[j]:=b[i][j];

  result:=0;
end;
//------------------------------------------------------------------------------
function THausholder.hausholder_2(FID2D:TFID_2D;progress:TProgressBar):integer;
(*
    Hausholder transformation̂ƁAɎOdΊpsJacobiőΊpI

    Y1 to n ƂBi0 to n-1ł͂ȂIj
    炩SetLength(n+1)ƂĂI

*)
var

  a,b,q:array of array of double;
  n:integer;
  u,p:array of double;

  l,k,j,i:integer;
  e,largeK, H, sqs:double;
  sa,sb,se,theta:double;
  c,s,t: double; // cos, sin, tan
  help1,help2:double;

//  motoTxt:string;

begin

 // motoTxt:=info.Text;

  n:=FID2D.al;

  setLength(a,n+1);
  setLength(b,n+1);
  setLength(q,n+1);

  for i:=0 to n do begin
    setLength(a[i],n+1);
    setLength(b[i],n+1);
    setLength(q[i],n+1);
  end;

  setLength(u,n+1);
  setLength(p,n+1);

  for i:=1 to n do for j:=1 to n do a[i][j]:=FID2D.FID[i-1].real.sig[j-1];


  //
  //
  //  Hausholder
  //
  //

 // info.Text:='Hausholder transformation...';
  for j:=1 to n do for k:=1 to n do if j=k then q[j][k]:=1.0 else q[j][k]:=0.0;
  //Pʍs


  progress.Max:=n;
  progress.Visible:=true;

  for i:=n downto 2 do begin

    progress.Position:=n-i;
    //
    //  sqrt(sigma)
    //
    sqs:=0.0;
    for k:=1 to i-1 do sqs:=sqs+a[k][i]*a[k][i];
    sqs:=sqrt(sqs);

    //
    //  u
    //
    for k:=1 to i-2 do u[k]:=a[k][i];

    if a[i-1][i]>=0 then begin
      u[i-1]:=a[i-1][i]+sqs;
      e:=-sqs;
    end else begin
      u[i-1]:=a[i-1][i]-sqs;
      e:=sqs;
    end; // e͕Ίpvfia[i-1][i]ɉl̕𔽓]́j

    for k:=i to n do u[k]:=0.0;

    //
    //  H
    //
    H:=0.0;
    for k:=1 to i-1 do H:=H+u[k]*u[k];
    H:=H/2.0;
    if H=0 then continue;  //Ń[ŊZȂ悤ɁBBB

    //
    //  p
    //
    for j:=1 to i do begin
      p[j]:=0.0;
      for k:=1 to i-1 do p[j]:=p[j]+a[j][k]*u[k];
      p[j]:=p[j]/H;
    end;

    //
    //  K(largeK)
    //
    largeK:=0.0;
    for k:=1 to i-1 do largeK:=largeK+u[k]*p[k];
    largeK:=largeK/H;

    //
    //   update a[j][k]
    //

    for j:=1 to i-1 do begin
      for k:=1 to i-1 do begin
        a[j][k]:=a[j][k]-p[j]*u[k]-u[j]*p[k]+largeK*u[j]*u[k];
      end;
    end; // j

    for k:=1 to i-2 do begin
      a[k][i]:=0.0;    //  [ɂȂ͂B
      a[i][k]:=0.0;    //  ->IɃ[I
    end;

    a[i-1][i]:= e;   // ɂ͒lciΊpvfj
    a[i][i-1]:= e;

    //
    // ssP=(1-u.u/H)uvɕt
    //
    for j:=1 to i-1 do begin
      for k:=1 to i-1 do begin
        b[j][k]:=0.0;
        for l:=1 to i do begin
          if j=l then b[j][k]:=b[j][k]+(1.0-u[j]*u[l]/H)*q[l][k]
          else b[j][k]:=b[j][k]+(0.0-u[j]*u[l]/H)*q[l][k];
        end;
      end;
    end;

    for j:=1 to i-1 do begin   // i-th row, i-th column͍XVĂ͂ȂȂI
      for k:=1 to i-1 do begin // loopi-1܂ŁI
        q[j][k]:=b[j][k];
      end;
    end;

  end; // i


  progress.max:=n-1;

  //
  // Jacobi]
  //
//  info.Text:='Jacobi rotation...';
  for i:=1 to n-1 do begin

    progress.Position:=i;

    se:=a[i][i+1];
    sa:=a[i][i];
    sb:=a[i+1][i+1];

    if se=0 then continue;
    theta:=0.5*(sb-sa)/se;

    if (abs(theta)+100*se=abs(theta)) then t:=se/(sb-sa)
    else begin
      t:=1/(abs(theta)+sqrt(1+theta*theta));
      if theta<0 then t:=-t;
    end;

    a[i][i]:=a[i][i]-t*se;
    a[i+1][i+1]:=a[i+1][i+1]+t*se;
    a[i][i+1]:=0;
    a[i+1][i]:=0;

    c:=1/(sqrt(1+t*t));
    s:=t*c;

    for k:=1 to n do begin
      help1:=c*q[i][k]-s*q[i+1][k];
      help2:=s*q[i][k]+c*q[i+1][k];
      q[i][k]:=help1;
      q[i+1][k]:=help2;
    end;

  end;

  //
  // Ίpvf̃[gI ̒l̓[ɂႤB
  //
  // info.Text:='Calculating square root...';
  for k:=1 to n do if a[k][k]>0 then a[k][k]:=sqrt(a[k][k]) else a[k][k]:=0;


  //
  // ----
  //
  // Ƃ̍svZɂ́AQ^{-1} A QvZB Q^{-1}Q̓]u
  // ܂Q^{-1} ǍvZB
  for i:=1 to n do for j:=1 to n do b[i][j]:=q[j][i]*a[j][j];
                                          // q]uI

  //ɉEQB
  for i:=1 to n do begin
  progress.Position:=i;
    for j:=1 to n do begin
      a[i][j]:=0.0;
      for k:=0 to n-1 do a[i][j]:=a[i][j]+b[i][k]*q[k][j];
                                  // E炩ꍇ[k][j]œ]uuȂvI
    end;
  end;

  //Q^{-1}QPʍsɂȂ邩ǂ`FbN
  for i:=1 to n do begin
    for j:=1 to n do begin
      b[i][j]:=0.0;
      for k:=0 to n-1 do b[i][j]:=b[i][j]+q[k][i]*q[k][j];
    end; // j
  end; // i

  for i:=1 to n do for j:=1 to n do begin
    FID2D.FID[i-1].real.sig[j-1]:=a[i][j];  // rooťʂrealɊi[
    FID2D.FID[i-1].imag.sig[j-1]:=b[i][j];  // ŗLxNgimagɊi[
    FID2D.FID[i-1].abs.sig[j-1]:=0;         // abs̓[ɂĂ
  end;
  //
  // ----
  //

//  info.Text:=motoTxt;
  progress.Visible:=false;

  result:=0;

end;
//------------------------------------------------------------------------------
function THausholder.hausholder_1(FID2D:TFID_2D):integer;
(*
    Y1 to n ƂBi0 to n-1ł͂ȂIj
    炩SetLength(n+1)ƂĂI

*)
var

  a,b,q:array of array of double;
  n:integer;
  u,p:array of double;

  l,k,j,i:integer;
  e,largeK, H, sqs:double;
begin

  n:=FID2D.al;

  setLength(a,n+1);
  setLength(b,n+1);
  setLength(q,n+1);

  for i:=0 to n do begin
    setLength(a[i],n+1);
    setLength(b[i],n+1);
    setLength(q[i],n+1);
  end;

  setLength(u,n+1);
  setLength(p,n+1);

  for i:=1 to n do for j:=1 to n do a[i][j]:=FID2D.FID[i-1].real.sig[j-1];


  //
  //
  //  Hausholder
  //
  //


  for j:=1 to n do for k:=1 to n do if j=k then q[j][k]:=1.0 else q[j][k]:=0.0;
  //Pʍs

  for i:=n downto 2 do begin

    //
    //  sqrt(sigma)
    //
    sqs:=0.0;
    for k:=1 to i-1 do sqs:=sqs+a[k][i]*a[k][i];
    sqs:=sqrt(sqs);

    //
    //  u
    //
    for k:=1 to i-2 do u[k]:=a[k][i];

    if a[i-1][i]>=0 then begin
      u[i-1]:=a[i-1][i]+sqs;
      e:=-sqs;
    end else begin
      u[i-1]:=a[i-1][i]-sqs;
      e:=sqs;
    end; // e͕Ίpvfia[i-1][i]ɉl̕𔽓]́j

    for k:=i to n do u[k]:=0.0;

    //
    //  H
    //
    H:=0.0;
    for k:=1 to i-1 do H:=H+u[k]*u[k];
    H:=H/2.0;
    if H=0 then continue;  //Ń[ŊZȂ悤ɁBBB

    //
    //  p
    //
    for j:=1 to i do begin
      p[j]:=0.0;
      for k:=1 to i-1 do p[j]:=p[j]+a[j][k]*u[k];
      p[j]:=p[j]/H;
    end;

    //
    //  K(largeK)
    //
    largeK:=0.0;
    for k:=1 to i-1 do largeK:=largeK+u[k]*p[k];
    largeK:=largeK/H;

    //
    //   update a[j][k]
    //

    for j:=1 to i-1 do begin
      for k:=1 to i-1 do begin
        a[j][k]:=a[j][k]-p[j]*u[k]-u[j]*p[k]+largeK*u[j]*u[k];
      end;
    end; // j

    for k:=1 to i-2 do begin
      a[k][i]:=0.0;    //  [ɂȂ͂B
      a[i][k]:=0.0;    //  ->IɃ[I
    end;

    a[i-1][i]:= e;   // ɂ͒lciΊpvfj
    a[i][i-1]:= e;

    //
    // ssP=(1-u.u/H)uvɕt
    //
    for j:=1 to i-1 do begin
      for k:=1 to i-1 do begin
        b[j][k]:=0.0;
        for l:=1 to i do begin
          if j=l then b[j][k]:=b[j][k]+(1.0-u[j]*u[l]/H)*q[l][k]
          else b[j][k]:=b[j][k]+(0.0-u[j]*u[l]/H)*q[l][k];
        end;
      end;
    end;

    for j:=1 to i-1 do begin   // i-th row, i-th column͍XVĂ͂ȂȂI
      for k:=1 to i-1 do begin // loopi-1܂ŁI
        q[j][k]:=b[j][k];
      end;
    end;

  end; // i

  //
  //
  //

  for i:=1 to n do for j:=1 to n do begin
    FID2D.FID[i-1].real.sig[j-1]:=a[i][j]; // OdΊpsrealɊi[
    FID2D.FID[i-1].imag.sig[j-1]:=q[i][j]; // ŗLxNgimagɊi[
    FID2D.FID[i-1].abs.sig[j-1]:=0.0; // abs̓[ɂĂ
  end;

  result:=0;

end;
//------------------------------------------------------------------------------
function THausholder.Process(FID2D: TFID_2D):integer;
begin
  //
  //  ŝݎ󂯕tBCovariance͕KȂĂ͂B
  //
  if FID2D.al<>FID2D.length_2D then begin
    error:=true;
    errorMessage:='Hausholder decomposition requires nxn matrix.';
    result:=1;
    exit;
  end;

  result:=hausholder_1(FID2D);

end;
//------------------------------------------------------------------------------
function TinverseHausholder.Process(FID2D: TFID_2D):integer;
var r:single;
    a:array of array of double;
    i,j,k,n:integer;
begin

  if FID2D.al<>FID2D.length_2D then begin
    error:=true;
    errorMessage:='Hausholder recomposition requires nxn matrix.';
    result:=1;
    exit;
  end;


  n:=FID2D.al;
  setLength(a,n);
  for k:=0 to n-1 do setLength(a[k],n);

  // Ƃ̍svZɂ́AQ^{-1} A QvZB Q^{-1}Q̓]u
  // ܂Q^{-1} ǍvZB
  for i:=0 to n-1 do begin
    for j:=0 to n-1 do begin
      r:=0.0;
      for k:=0 to n-1 do r:=r+FID2D.FID[k].imag.sig[i]*FID2D.FID[k].real.sig[j];
                                  // [i][k]->[k][i] œ]uI
      a[i][j]:=r;
    end; // j
  end; // i

  //ɉEQB
  for i:=0 to n-1 do begin
    for j:=0 to n-1 do begin
      r:=0.0;
      for k:=0 to n-1 do r:=r+a[i][k]*FID2D.FID[k].imag.sig[j];
                                  // E炩ꍇ[k][j]œ]uuȂvI
      FID2D.FID[i].real.sig[j] := r;
    end;
  end;

  //Q^{-1}QPʍsɂȂ邩ǂ`FbN
  for i:=0 to n-1 do begin
    for j:=0 to n-1 do begin
      a[i][j]:=0.0;
      for k:=0 to n-1 do
        a[i][j]:=a[i][j]+FID2D.FID[k].imag.sig[i]*FID2D.FID[k].imag.sig[j];
    end; // j
  end; // i

  for i:=0 to n-1 do for j:=0 to n-1 do FID2D.FID[i].imag.sig[j]:=a[i][j];


  result:=0;

end;

//------------------------------------------------------------------------------
function TCrossCorrelation.Process(FID2D: TFID_2D):integer;
var r: array of array of double;
    i,j:cardinal;
begin
  //
  //  ŝݎ󂯕tBCovariance͕KȂĂ͂B
  //
  if FID2D.al<>FID2D.length_2D then begin
    error:=true;
    errorMessage:='Cross correlation requires the data to be nxn matrix.';
    result:=1;
    exit;
  end;

  setLength(r,FID2D.al);
  for i:=0 to FID2D.length_2D-1 do setLength(r[i],FID2D.length_2D);
  for i:=0 to FID2D.length_2D-1 do for j:=0 to FID2D.length_2D-1 do
    r[i][j]:= FID2D.FID[i].real.sig[j]
                / sqrt(FID2D.FID[i].real.sig[i] * FID2D.FID[j].real.sig[j]);

  for i:=0 to FID2D.length_2D-1 do for j:=0 to FID2D.length_2D-1 do begin
    FID2D.FID[i].real.sig[j] := r[i][j];
    FID2D.FID[i].imag.sig[j] := 0.0;
    FID2D.FID[i].abs.sig[j] := r[i][j];

  end;

  error:=false;
  errorMessage:='';
  result:=0;

end;
//------------------------------------------------------------------------------
function TTRanspose.Process(FID2D: TFID_2D):integer;
var i,j:integer;
    newAL,newLength_2D:cardinal;
    cr,ci,ca: array of array of double;
begin
  //
  //  c?[i][j]̃m
  //
  SetLength(cr,FID2D.al);
  for i:=0 to FID2D.al-1 do SetLength(cr[i],FID2D.length_2D);
  SetLength(ci,FID2D.al);
  for i:=0 to FID2D.al-1 do SetLength(ci[i],FID2D.length_2D);
  SetLength(ca,FID2D.al);
  for i:=0 to FID2D.al-1 do SetLength(ca[i],FID2D.length_2D);

  //
  //  c?[i][j]̏
  //
  for i:=0 to FID2D.al-1 do for j:=0 to FID2D.length_2D-1 do begin
    cr[i][j]:=FID2D.FID[j].real.sig[i];
    ci[i][j]:=FID2D.FID[j].imag.sig[i];
    ca[i][j]:=FID2D.FID[j].abs.sig[i];
  end;

  newAL:=FID2D.length_2D;
  newLength_2D:=FID2D.al;
  //
  // FID2D array lengthƂ肠P
  //
  while FID2D.length_2D>1 do begin
    FID2D.delete(FID2D.length_2D-1);
  end;

  //
  // FID2DALƂƂ
  //
  FID2D.al:=newAL;

  if FID2D.length_2D<newLength_2D then
    repeat FID2D.add; until FID2D.length_2D=newLength_2D;

  for i:=0 to FID2D.length_2D-1 do begin
    for j:=0 to FID2D.al-1 do begin
      FID2D.FID[i].real.sig[j]:=cr[i][j];
      FID2D.FID[i].imag.sig[j]:=ci[i][j];
      FID2D.FID[i].abs.sig[j]:=ca[i][j];
    end;
  end;

  error:=false;
  errorMessage:='';
  result:=0;

end;

//------------------------------------------------------------------------------
function TCovariance.realCovariance(FID2D:TFID_2D):integer;
var i,j,k:integer;
    a: double;
    c: array of array of double;
    s: array of double;
begin
  //
  //  c[i][j]̃m
  //
  SetLength(c,FID2D.al);
  for j:=0 to FID2D.al-1 do SetLength(c[j],FID2D.al);

  //
  //  c[i][j]̏
  //
  for i:=0 to FID2D.al-1 do for j:=0 to FID2D.al-1 do c[i][j]:=0.0;


  //
  //  s[j]ic̕ρj̏ƌvZ
  //
  setLength(s,FID2D.al);
  for i:=0 to FID2D.al-1 do s[i]:=0.0;

  if not innerProductQ then begin
   // innerProductQ=trueȂXLbv-> s̓[̂܂܁B

  for j:=0 to FID2D.al-1 do begin
    for i:=0 to FID2D.length_2D-1 do begin
      s[j]:=s[j]+FID2D.FID[i].real.sig[j];
    end; //i
    s[j]:=s[j]/FID2D.length_2D;
  end; //j

  end; // if not innerProductQ;

  for i:=0 to FID2D.al-1 do begin
    for j:=0 to FID2D.al-1 do begin

      a:=0.0;
      for k:=0 to FID2D.length_2D-1 do begin
        a:=a+(FID2D.FID[k].real.sig[i]-s[i])*(FID2D.FID[k].real.sig[j]-s[j]);
      end;
      c[i][j]:= a/(FID2D.length_2D-1);

    end; // j
  end; //i


  //
  //  c[i][j] -> FID2DRealɑ
  //
  if FID2D.length_2D<FID2D.al then
    repeat FID2D.add; until FID2D.length_2D=FID2D.al
  else if FID2D.length_2D>FID2D.al then
    repeat FID2D.delete(FID2D.length_2D-1); until FID2D.length_2D=FID2D.al;

  for i:=0 to FID2D.al-1 do for j:=0 to FID2D.al-1 do FID2D.FID[i].real.sig[j]:=c[i][j];

  error:=false;
  errorMessage:='';
  result:=0;
end;
//------------------------------------------------------------------------------
function TCovariance.hyperComplexCovariance(FID2D:TFID_2D):integer;
var i,j,k:integer;
    ar,ai: double;
    p,q,r,s: double;
    cr,ci: array of array of double;
    sr,si: array of double;
begin

  if (FID2D.length_2D mod 2) <>0 then begin
    error:=true;
    errorMessage:='Covariance: In the hypercomplex mode, array length should be even.';
    result:=1;
    exit;
  end;

  //
  //  cr[i][j], ci[i][j]̃m
  //
  SetLength(cr,FID2D.al);
  SetLength(ci,FID2D.al);
  for j:=0 to FID2D.al-1 do begin
    SetLength(cr[j],FID2D.al);
    SetLength(ci[j],FID2D.al);
  end;

  //
  //  cr[i][j], ci[i][j]̏
  //
  for i:=0 to FID2D.al-1 do for j:=0 to FID2D.al-1 do cr[i][j]:=0.0;
  for i:=0 to FID2D.al-1 do for j:=0 to FID2D.al-1 do ci[i][j]:=0.0;


  //
  //  sr[j], si[j]ic̕ρj̏ƌvZ
  //  ЂƂre,im,re,im,...
  //
  setLength(sr,FID2D.al);
  setLength(si,FID2D.al);
  for i:=0 to FID2D.al-1 do begin sr[i]:=0.0; si[i]:=0.0; end;

  if not innerProductQ then begin
   // innerProductQ=trueȂXLbv-> sr,si̓[̂܂܁B

  for j:=0 to FID2D.al-1 do begin
    i:=0;
    while i<integer(FID2D.length_2D) do begin
      sr[j]:=sr[j]+FID2D.FID[i].real.sig[j];
      inc(i);
      si[j]:=si[j]+FID2D.FID[i].real.sig[j];
      inc(i);
    end; // while

    sr[j]:=2.0*sr[j]/FID2D.length_2D;
    si[j]:=2.0*si[j]/FID2D.length_2D;
  end; //j

  end; // if not innerProductQ


  for i:=0 to FID2D.al-1 do begin
    for j:=0 to FID2D.al-1 do begin

      ar:=0.0; ai:=0.0;

      k:=0;
      while k<integer(FID2D.length_2D-1) do begin
        p:=FID2D.FID[k].real.sig[i]-sr[i];
        q:=FID2D.FID[k+1].real.sig[i]-si[i];
        r:=FID2D.FID[k].real.sig[j]-sr[j];
        s:=FID2D.FID[k+1].real.sig[j]-si[j];

        ar:=ar+(p*r+q*s);
        ai:=ai+(q*r-p*s);

        k:=k+2;
      end; // while k

      cr[i][j]:= ar/((FID2D.length_2D div 2) -1 );
      ci[i][j]:= ai/((FID2D.length_2D div 2) -1 );

    end; // j
  end; //i



  //
  //  cr[i][j] -> FID2DRealɑ ci[i][j] -> FID2Dimagɑ
  //

  if FID2D.length_2D<FID2D.al then
    repeat FID2D.add; until FID2D.length_2D=FID2D.al
  else if FID2D.length_2D>FID2D.al then
    repeat FID2D.delete(FID2D.length_2D-1); until FID2D.length_2D=FID2D.al;


  for i:=0 to FID2D.al-1 do for j:=0 to FID2D.al-1 do begin

    FID2D.FID[i].real.sig[j]:=cr[i][j];
    FID2D.FID[i].imag.sig[j]:=ci[i][j];
    FID2D.FID[i].abs.sig[j]:=sqrt(cr[i][j]*cr[i][j]+ci[i][j]*ci[i][j]);

  end;

  error:=false;
  errorMessage:='';
  result:=0;
end;
//------------------------------------------------------------------------------
function TCovariance.Process(FID2D:TFID_2D):integer;
begin

  error:=false;
  errorMessage:='';

  if FID2D.length_2D<2 then begin
    error:=true;
    errorMessage:='Covariance cannot be applied to 1D data.';
    result:=1;
    exit;
  end;

  if FID2D.al<2 then begin
    error:=true;
    errorMessage:='Covariance: al should be at least 2.';
    result:=1;
    exit;
  end;

  if covarianceMode=cvReal then realCovariance(FID2D)
  else if covarianceMode=cvHyperComplex then hyperComplexCovariance(FID2D);

  if error then begin
    result:=1;
    exit;
  end;

  result:=0;

end;
//------------------------------------------------------------------------------
function TPivot.Process(FID2D:TFID_2D):integer;
var k:cardinal;
begin
  for k:=0 to FID2D.length_2D-1 do begin
    FID2D.FID[k].pivot:=point;
  end;
  FID2D.pivot:=point;
  result:=0;
end;

function TPivot.Process(FID:TFID):integer;
begin
  FID.pivot:=point;
  result:=0;
end;
//------------------------------------------------------------------------------
function TFIDInverseFFT.Process(FID2D:TFID_2D):integer;
var k: cardinal;
begin
  for k:=0 to FID2D.length_2D-1 do begin
    Process(FID2D.FID[k]);
  end;
  result:=0;
end;

//------------------------------------------------------------------------------
function TFIDInverseFFT.Process(FID:TFID):integer;
var
  theta:double;
  yr,yi,xr,xi,wr,wi: double;
  i,j,k,m,mh,irev,n,nh: cardinal;
begin

with FID do begin
  // FAL͂Q̉?
  n:=1;
  while n<AL do begin
    n := n shl 1;
  end; // while

  // FALQnłȂ΁A[tB
  if AL<n then begin
    k:=AL;
    AL:=n;
    setLength(real.sig,AL);
    setLength(imag.sig,AL);
    setLength(abs.sig,AL);
    for m:=k to AL-1 do begin
      real.sig[m]:=0.0;
      imag.sig[m]:=0.0;
    end; // for

    warning:=true;
    warningMessage:='Warning: The data was automatically zero-filled to '
        + IntToStr(AL) +' points.';
  end;

  theta:=6.28318530717958647692528676656/n;

  // -----------  inverse FFT -------------

  // STEP 1: NMRp̕בւ
  //    iS[ŁAvXgE}CiXgɂ̂߂Bj
  nh:=n div 2;
  for k:=0 to nh-1 do begin
    xr:=real.sig[k]; xi:=imag.sig[k];
    real.sig[k]:=real.sig[nh+k]; imag.sig[k]:=imag.sig[nh+k];
    real.sig[nh+k]:=xr; imag.sig[nh+k]:=xi;
  end;


  // STEP 2
  m:=n;
  mh:=m shr 1;

  while m>=2 do begin
    irev:=0;
    i:=0;
    while i<n do begin
      wr:=cos(theta * irev);
      wi:=sin(theta * irev);

      k:=n shr 2;
      irev:=irev xor k;
      while k>irev do begin
        k:=k shr 1;
        irev:=irev xor k;
      end; //while

      for j:=i to mh+i-1 do begin
        k:=j+mh;
        xr:=wr*real.sig[k]+wi*imag.sig[k];   // FFT̂Ƃ+-tI
        xi:=wr*imag.sig[k]-wi*real.sig[k];   // I
        yr:=real.sig[j]; yi:=imag.sig[j];
        real.sig[j]:=yr+xr;
        imag.sig[j]:=yi+xi;
        real.sig[k]:=yr-xr;
        imag.sig[k]:=yi-xi;;
      end; // for
    i:=i+m;
    end; // while
  m:=mh;
  mh:=mh shr 1;
  end; // while


  //STEP 3
  i:=0;
  for j:=1 to n-2 do begin
    k:=n shr 1;
    i:=i xor k;
    while k>i do begin
      k:=k shr 1;
      i:=i xor k;
    end; // while
    if j<i then begin
      xr:=real.sig[j];
      xi:=imag.sig[j];
      real.sig[j]:=real.sig[i];
      imag.sig[j]:=imag.sig[i];
      real.sig[i]:=xr;
      imag.sig[i]:=xi;
    end; // if(j<i)

  end; // for(j)


  // Abs ̌vZ
  for k:=0 to AL-1 do begin
    real.sig[k]:=real.sig[k];
    imag.sig[k]:=imag.sig[k];
    abs.sig[k]:=sqrt(real.sig[k]*real.sig[k] + imag.sig[k]*imag.sig[k]);
  end;

  if domain=dmTime then domain:=dmFrequency
  else if domain=dmFrequency then domain:=dmTime;



end; // with FID
 result:=0;
end;

//------------------------------------------------------------------------------
function TFIDILT.Process(FID2D:TFID_2D):integer;
var k:integer;
begin
  error:=false;
  for k:=0 to FID2D.length_2D-1 do begin
    if Process(FID2D.FID[k])<>0 then error:=true;
  end;
  if error then result:=1 else result:=0;
end;
//------------------------------------------------------------------------------
function TFIDILT.Process(FID: TFID):integer;
var
  n:cardinal;
  ex: double;
  proc:TFIDInverseFFT;
begin

  proc:=TFIDInverseFFT.Create;
  error:=false;
  if proc.Process(FID)<>0 then begin
    error:=true;
    errorMessage:=proc.errorMessage;
    proc.Free;
    result:=-1;
    exit;
  end;


  for n:=0 to FID.AL-1 do begin
    ex:=exp(FA*n/FID.AL);
    FID.real.sig[n]:=FID.real.sig[n]*ex;
    FID.imag.sig[n]:=FID.imag.sig[n]*ex;
    FID.abs.sig[n]:=FID.abs.sig[n]*ex;
  end;

  proc.Free;
  result:=0;


end;
//------------------------------------------------------------------------------
function TFIDLT.Process(FID2D:TFID_2D):integer;
var k:integer;
begin
  error:=false;
  for k:=0 to FID2D.length_2D-1 do begin
    if Process(FID2D.FID[k])<>0 then error:=true;
  end;
  if error then result:=1 else result:=0;
end;
//------------------------------------------------------------------------------
function TFIDLT.Process(FID: TFID):integer;
var
  n:cardinal;
  ex:double;
  proc:TFIDFFT;
begin
  for n:=0 to FID.AL-1 do begin
    ex:=exp(-1.0*FA*n/FID.AL);
    FID.real.sig[n]:=FID.real.sig[n]*ex;
    FID.imag.sig[n]:=FID.imag.sig[n]*ex;
  end;

  proc:=TFIDFFT.Create;
  if proc.Process(FID)<>0 then begin
    error:=true;
    errorMessage:=proc.errorMessage;
    result:=-1;
    proc.Free;
    exit;
  end;

  proc.Free;
  result:=0;

end;
//------------------------------------------------------------------------------
function TFIDFFT.Process(FID2D:TFID_2D):integer;
var k: cardinal;
begin
  for k:=0 to FID2D.length_2D-1 do begin
    Process(FID2D.FID[k]);
  end;
  result:=0;
end;
//------------------------------------------------------------------------------
function TFIDFFT.Process(FID:TFID):integer;
var
  theta:double;
  xr,xi,wr,wi: double;
  i,j,k,m,mh,irev,n,nh: cardinal;
begin

with FID do begin
  // FAL͂Q̉?
  n:=1;
  while n<AL do begin
    n := n shl 1;
  end; // while

  // FALQnłȂ΁A[tB
  if AL<n then begin
    k:=AL;
    AL:=n;
    setLength(real.sig,AL);
    setLength(imag.sig,AL);
    setLength(abs.sig,AL);
    for m:=k to AL-1 do begin
      real.sig[m]:=0.0;
      imag.sig[m]:=0.0;
      abs.sig[m]:=0.0;
    end; // for

    warning:=true;
    warningMessage:='Warning: The data was automatically zero-filled to '
        + IntToStr(AL) +' points.';
  end;

  theta:=6.28318530717958647692528676656/n;

  // -----------  FFT -------------

  i:=0;
  for j:=1 to n-2 do begin
    k:=n shr 1;
    i:=i xor k;
    while k>i do begin
      k:=k shr 1;
      i:=i xor k;
    end; // while
    if j<i then begin
      xr:=real.sig[j];
      xi:=imag.sig[j];
      real.sig[j]:=real.sig[i];
      imag.sig[j]:=imag.sig[i];
      real.sig[i]:=xr;
      imag.sig[i]:=xi;
    end; // if(j<i)

  end; // for(j)


  mh:=1;
  m:=mh shl 1;
  while m<=n do begin
    irev:=0;
    i:=0;
    while i<n do begin
      wr:=cos(theta * irev);
      wi:=sin(theta * irev);

      k:=n shr 2;
      irev:=irev xor k;
      while k>irev do begin
        k:=k shr 1;
        irev:=irev xor k;
      end; //while

      for j:=i to mh+i-1 do begin
        k:=j+mh;
        xr:=real.sig[j]-real.sig[k];
        xi:=imag.sig[j]-imag.sig[k];
        real.sig[j]:=real.sig[j]+real.sig[k];
        imag.sig[j]:=imag.sig[j]+imag.sig[k];
        real.sig[k]:=wr*xr-wi*xi;
        imag.sig[k]:=wr*xi+wi*xr;
      end; // for
    i:=i+m;
    end; // while
  mh:=m;
  m:=mh shl 1;
  end; // while

  // NMRp̕בւiS[ŁAvXgE}CiXgj
  nh:=n div 2;
  for k:=0 to nh-1 do begin
    xr:=real.sig[k]; xi:=imag.sig[k];
    real.sig[k]:=real.sig[nh+k]; imag.sig[k]:=imag.sig[nh+k];
    real.sig[nh+k]:=xr; imag.sig[nh+k]:=xi;
  end;


  // Abs ̌vZ
  for k:=0 to AL-1 do begin
    real.sig[k]:=real.sig[k]/AL;
    imag.sig[k]:=imag.sig[k]/AL;
    abs.sig[k]:=sqrt(real.sig[k]*real.sig[k] + imag.sig[k]*imag.sig[k]);
  end;


  if domain=dmTime then domain:=dmFrequency
  else if domain=dmFrequency then domain:=dmTime;

end; // with
result:=0;
end;

//------------------------------------------------------------------------------

function TFIDZeroFill.Process(FID:TFID):integer;
var
  preval:cardinal;
  k:cardinal;
begin
  preval:=FID.AL;

  if preval+points>65536 then begin
    result:=1;
    error:=true;
    errormessage:='Zero-filling: Too many points.';
    exit;
  end;

  FID.AL:=preval+points;

  if zeroFillMode=zfAppend then begin
    while preval<FID.AL do begin
      FID.real.sig[preval]:=0.0;
      FID.imag.sig[preval]:=0.0;
      FID.abs.sig[preval]:=0.0;
      inc(preval);
    end;
  end else begin // zfPrepend
    for k:=preval-1 downto 0 do begin
      FID.real.sig[k+points]:=FID.real.sig[k];
      FID.imag.sig[k+points]:=FID.imag.sig[k];
      FID.abs.sig[k+points]:=FID.abs.sig[k];
    end; // k
    for k:=points-1 downto 0 do begin
      FID.real.sig[k]:=0.0;
      FID.imag.sig[k]:=0.0;
      FID.abs.sig[k]:=0.0;
    end;

  end;
  result:=0;


end;

function TFIDZeroFill.Process(FID2D:TFID_2D):integer;
var k:integer;
begin
  for k:=0 to FID2D.length_2D-1 do begin
    Process(FID2D.FID[k]);
  end; // for
  FID2D.al:=FID2D.FID[0].al;
  if error then result:=1 else result:=0;
end;
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
function TFIDPhase1.Process(FID:TFID):integer;
var
  phi,dphi,c,s: double;
  xr,xi: double;
  p,i: integer;
begin

  p:=Integer(FID.pivot);

  phi:=3.14159265358979*angle*(p/FID.AL)/180.0;
  dphi:=-3.14159265358979*angle*(1/FID.AL)/180.0;

  i:=0;
  while i<integer(FID.AL) do begin

//    phi:=3.14159265358979*angle*((p-i)/FID.AL)/180.0;
         // degree -> radian conversion
    c:=cos(phi);
    s:=sin(phi);

    xr:=FID.real.sig[i];
    xi:=FID.imag.sig[i];
    FID.real.sig[i] := c*xr - s*xi;
    FID.imag.sig[i] := c*xi + s*xr;

    phi:=phi+dphi;

    inc(i);
  end; // i

  result:=0;
end;

function TFIDPhase1.Process(FID2D:TFID_2D):integer;
var k,res:integer;
begin
  res:=0;

  for k:=0 to FID2D.length_2D-1 do begin
    res:=Process(FID2D.FID[k]);
  end;
  result:=res;

end;
//------------------------------------------------------------------------------
function TFIDPhase0.Process(FID:TFID):integer;
begin
  FID.rotate(angle);
  error:=false;
  errorMessage:='';
  result:=0;
end;
//------------------------------------------------------------------------------
function TFIDPhase0.Process(FID2D:TFID_2D):integer;
var j:cardinal;
begin
  for j:=0 to FID2D.length_2D-1 do Process(FID2D.FID[j]);
  error:=false;
  errorMessage:='';
  result:=0;
end;
//------------------------------------------------------------------------------
function TFIDCutTail.Process(FID:TFID):integer;
begin
  if points>FID.AL-1 then result:=1 else begin
    FID.AL:=FID.AL-points;
    error:=false;
    errorMessage:='';
    result:=0;
  end;
end;

function TFIDCutTail.Process(FID2D:TFID_2D):integer;
var j:cardinal;
begin
  error:=false;
  errorMessage:='';

  for j:=0 to FID2D.length_2D-1 do begin
   if Process(FID2D.FID[j])<>0 then error:=true;
  end;

  if error then begin
    errorMessage:='Cut-tail: You are trying to cut more points than possible.';
    result:=1;
  end else begin
    FID2D.AL:=FID2D.AL-points;
    result:=0;
  end;
end;
//------------------------------------------------------------------------------
function TFIDCutHead.Process(FID:TFID):integer;
var k: cardinal;
begin

  if points>FID.AL-1 then result:=1 else begin

    for k:=0 to FID.AL-1-points do begin
      FID.real.sig[k]:=FID.real.sig[k+points];
      FID.imag.sig[k]:=FID.imag.sig[k+points];
      FID.abs.sig[k]:=FID.abs.sig[k+points];
    end;

    FID.AL:=FID.AL-points;
    error:=false;
    errorMessage:='';
    result:=0;
  end;

end;

function TFIDCutHead.Process(FID2D:TFID_2D):integer;
var j:cardinal;
begin

  error:=false;
  errorMessage:='';

  for j:=0 to FID2D.length_2D-1 do begin
    if Process(FID2D.FID[j])<>0 then error:=true;
  end;

  if error then begin
    errorMessage:='Cut-head: You are trying to cut more points than possible.';
    result:=1;
  end else begin
    FID2D.AL:=FID2D.AL-points;
    result:=0;
  end;
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//------------------    CAPTIONS    --------------------------------------------
//------------------------------------------------------------------------------

function TFIDDivideScalar.caption:string;
begin
  caption:='Divide by ' + FloatToStr(FScalar);
end;
//------------------------------------------------------------------------------

function TFIDmultiplyScalar.caption:string;
begin
  caption:='Multiplied by ' + FloatToStr(FScalar);
end;
//------------------------------------------------------------------------------

function TFIDSubtractFID.caption:string;
begin
  caption:='Subtract by data file '+FDatFilename +
          ' with a weighting factor of ' + FloatToStr(FWeight);
end;

//------------------------------------------------------------------------------

function TFIDAddFID.caption:string;
begin
  caption:='Add by data file '+FDatFilename +
          ' with a weighting factor of ' + FloatToStr(FWeight);
end;
//------------------------------------------------------------------------------

function TFIDJoinFID.caption:string;
begin
  caption:='Join with data file '+FDatFilename +
          ' with a weighting factor of ' + FloatToStr(FWeight);
end;
//------------------------------------------------------------------------------

function TFIDMultiplyFID.caption:string;
begin
  caption:='Multiply with data file '+FDatFilename;
end;

//------------------------------------------------------------------------------

function TFIDDivideFID.caption:string;
begin
  caption:='Divide by data file '+FDatFilename;
end;
//------------------------------------------------------------------------------

function TFIDLT.caption:string;
begin
  caption:='Laplace transform: a='+FloatToStr(FA);
end;
//------------------------------------------------------------------------------

function TFIDILT.caption:string;
begin
  caption:='Inverse Laplace transform: a='+FloatToStr(FA);
end;
//------------------------------------------------------------------------------
function TFIDlinearBaselineCorrection.caption:string;
begin
  caption:='linear baseline correction: ('
          + IntToStr(x1) + ',' + FloatToStrF(y1Re,ffGeneral,5,5)
          +                ',' + FloatToStrF(y1Im,ffGeneral,5,5) + ') - ('
          + IntToStr(x2) + ',' + FloatToStrF(y2Re,ffGeneral,5,5)
          +                ',' + FloatToStrF(y2Im,ffGeneral,5,5) + ')'

end;
//------------------------------------------------------------------------------
function TFIDdcOffsetCut.caption:string;
begin
  caption:= 'DC-offset cut: subtraction by average of the last '
      + IntToStr(points) + ' points';
end;
//------------------------------------------------------------------------------
function TFIDPhase0.caption:string;
begin
  caption:= '0th-order phase rotation: ' + FloatToStr(angle) + ' degree';
end;
//------------------------------------------------------------------------------
function TFIDPhase1.caption:string;
begin
  caption:= '1st-order phase rotation: ' + FloatToStr(angle) + ' degree';
end;
{----------------------------------------------------------}
function TFIDCutHead.caption:string;
begin
  caption:= 'Cut first ' + IntToStr(points) + ' point(s)';
end;
{----------------------------------------------------------}
function TFIDCutTail.caption:string;
begin
  caption:= 'Cut last ' + IntToStr(points) + ' point(s)';
end;
//------------------------------------------------------------------------------
function TFIDZeroFill.caption:string;
begin
  caption:='Zero fill ' + IntToStr(points) + ' point(s)';
end;
//------------------------------------------------------------------------------
function TFIDFFT.caption:string;
begin
  if warning=false then
    caption:='Fast Fourier transformation (FFT)'
  else
    caption:='[FFT] ' + warningMessage;
end;
//------------------------------------------------------------------------------
function TFIDInverseFFT.caption:string;
begin
  if warning=false then
    caption:='Inverse Fast Fourier transformation (IFFT)'
  else
    caption:='[Inverse FFT] ' + warningMessage;
end;
//------------------------------------------------------------------------------
function TPivot.caption:string;
begin
  caption:='Set phase pivot at '+ IntToStr(point) + '-th point'
end;
//------------------------------------------------------------------------------
function TFIDApodization.caption:string;
var s:string;
begin
  case apType of
    Gaussian: s:='Gaussian';
    Lorenzian: s:='Lorenzian';
  end;

  caption:= s+' '+'apodization (' + FloatToStr(width) +'Hz)';
end;
//------------------------------------------------------------------------------
function TPCW.caption:string;
begin
  caption:='Phase Covariance Weighting (PCW)';
end;
//------------------------------------------------------------------------------
function TaverageOverT1.caption:string;
begin
  caption:='average over T1';
end;
//------------------------------------------------------------------------------
function TTranspose.caption:string;
begin
  caption:='Transpose';
end;
//------------------------------------------------------------------------------
function TFlatten.caption:string;
begin
  caption:='Flatten';
end;
//------------------------------------------------------------------------------
function TCovariance.caption:string;
var s:string;
begin
  case covarianceMode of
    cvReal: s:='(t1 real)';
    cvHyperComplex: s:='(t1 hypercomplex)';
  end;
  caption:='Covariance ' + s;
end;
//------------------------------------------------------------------------------
function TCrossCorrelation.caption:string;
begin
  caption:='Cross correlation';
end;
//------------------------------------------------------------------------------
function THausholder.caption:string;
begin
  caption:='Hausholder decomposition (Re <- diag. elements, Im <- trans matrix)';
end;
//------------------------------------------------------------------------------
function TinverseHausholder.caption:string;
begin
  caption:='Hausholder recomposition (for testing purpose)';
end;
//------------------------------------------------------------------------------
function TFIDExtrapolate.caption:string;
begin
  caption:='Extrapolate FID with '
          + IntToStr(FaddPoints) + ' points by average of the last '
          + IntToStr(FaveragePoints) + ' points.';
end;
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
//




end.
